#!/usr/bin/perl -w
# vim:ts=4
#
# Gather data for a vmware farm of servers.
#
# Steve Shipway, University of Auckland, 2007
#
# This is the new version for ESX3 that is Farm-aware

use strict;
use Getopt::Std;
use vars qw/$opt_g $opt_x $opt_C $opt_D $opt_B $opt_R $opt_h $opt_d $opt_t $opt_F/;
use RRDs;
use Net::SNMP;

#############################################################################
# Globals
my($MINVER) = 2.5;
my($UCDOID) = "1.3.6.1.4.1.2021.1000.10"; # UDC snmp daemon 
my($COMMUNITY,$FARM,$RRD,$MRTG,$DEBUG,$TIMEOUT)
	=('public','esx','/usr/local/rrdtool/esx','/usr/local/mrtg/conf',0,10);
my($SUBDIR)='esx';
my($DOMAIN)=".auckland.ac.nz";
my(%vmdata) = ();
my(%srvdata) = ();
my(%cache) = ();
my(%stats) = ();
my($HOST,$GUEST);
my($MAXID) = 0;
#############################################################################
# Utility subroutines
sub dohelp {
	print "Usage: $0 -F farm [-C community] -R rrddir -D mrtgcfgpath\n";
	print "       [-B subdir] [-t timeout] [-d] [-h] [-x] [-g] esxserverhostnames...\n";
	print "-x : Erase all .conf files first.\n";
	print "-d : Debug mode.\n";
	print "-g : Do not generate config files for individual guests.\n";
	print "-R : RRD files will live in this directory with farmname appended\n";
	print "-D : This is the base directory for MRTG config files\n";
	print "-B : MRTG config files live in base directory, plus this.\n";
	print "\n*NOTE* Requires RRDTool perl module to be installed.\n\n";
	print "This will create and update all the .rrd files for the listed servers.\n";
	print "It will also create dummy MRTG .conf files to allow this data to be displayed\n";
	print "using some MRTG frontend, eg routers2.cgi.  This will include summary conf\n";
	print "files for per-ESXserver and per-farm.\n";
	print "MRTG files are located in separate subdirs for farm and server.\n";
	print "Recommended routers2.cgi settings:\n";
	print "confpath=mrtgcfgpath\ncfgfiles=subdir/*.conf subdir/*/*.conf subdir/*/guests/*.conf\nfor specified mrtgcfgpath and subdir\n";
	exit 0;
}
sub readcache {
	my($file) = "$RRD/cache.$FARM";
	my($i,$k,$v,$n);
	return if(!-f $file);	
	if(!-r $file) {
		print "Error: cannot read cache $file\n";
		exit 1;
	}
	print "Reading cache file for $FARM...\n" if($DEBUG);
	%cache = (); # clean everything out
	open CACHE,"<$file" or do {
		print "Error: cannot open cache $file: $!\n";
		exit 1;
	};
	while ( <CACHE> ) {
		chomp;
		if( /^\s*(\S.*\S)\s*=\s*(.*\S)/ ) {
			($k,$v)=($1,$2);
			$cache{$k} = $v;
			if($k =~ /^guest-(.+)/) {
				$vmdata{$v}{name} = $1;
				$vmdata{$v}{state} = -1; # no longer exists?
			} elsif( $k =~ /^(\d+)-maxmem$/ ) {
				$vmdata{$1}{maxmem} = $v;
			} elsif( $k =~ /^(\d+)-maxcpu$/ ) {
				$vmdata{$1}{maxcpu} = $v;
			} elsif( $k =~ /^(\d+)-lastseen$/ ) {
				$vmdata{$1}{lastseen} = $v;
			} elsif( $k =~ /^srv-(\S+)-maxmem$/ ) {
				$srvdata{$1}{maxmem} = $v;
			} elsif( $k =~ /^srv-(\S+)-maxcpu$/ ) {
				$srvdata{$1}{maxcpu} = $v;
			}
		}
	}
	close CACHE;
	$MAXID = $cache{maxid} if($cache{maxid});
	print "...done\n" if($DEBUG);
}
sub writecache {
	my($file) = "$RRD/cache.$FARM";
	print "Updating cache file for $FARM...\n" if($DEBUG);
	$cache{maxid} = $MAXID;
	open CACHE,">$file" or do {
		print "Error: cannot open cache $file: $!\n";
		exit 1;
	};
	foreach ( keys %cache ) { 
		print CACHE "$_ = ".$cache{$_}."\n" if($_ and defined $cache{$_}); }
	close CACHE;
}
sub getid($) {
	# get ID for this Guest.  Return 0 if we dont know it yet.
	my($guest) = $_[0];
	my($id);
	$id = $cache{"guest-$guest"};
	if($id) {
		# read cache
		$vmdata{$id}{maxmem} = $cache{"$id-maxmem"} if($cache{"$id-maxmem"});
		$vmdata{$id}{maxcpu} = $cache{"$id-maxcpu"} if($cache{"$id-maxcpu"});
		return $id;
	}
	return 0;
}
sub addguest($) {
	my($guest) = $_[0];
	my($rrd,$err);
	my($id);
	# Add a new guest with this description.
	# increment maxid
	$id = $MAXID + 1; 
	# Create .rrd files
	$rrd = "$RRD/$FARM/$id";
	print "Creating RRD files for '$guest' ($id)\n" if($DEBUG);
	RRDs::create( "$rrd-cpu.rrd", # used/sys
		qw/RRA:AVERAGE:0.5:1:800 RRA:AVERAGE:0.25:6:800 RRA:AVERAGE:0.25:24:800 RRA:AVERAGE:0.25:288:800 RRA:MAX:0.5:1:800 RRA:MAX:0.25:6:800 RRA:MAX:0.25:24:800 RRA:MAX:0.25:288:800/,
		qw/DS:ds0:COUNTER:600:0:32 DS:ds1:COUNTER:600:0:32/ );
	$err = RRDs::error; if($err) { print "$id-cpu RRD Error: $err\n"; return -1; }
	RRDs::create( "$rrd-rdy.rrd", # ready
		qw/RRA:AVERAGE:0.5:1:800 RRA:AVERAGE:0.25:6:800 RRA:AVERAGE:0.25:24:800 RRA:AVERAGE:0.25:288:800 RRA:MAX:0.5:1:800 RRA:MAX:0.25:6:800 RRA:MAX:0.25:24:800 RRA:MAX:0.25:288:800/,
		qw/DS:ds0:COUNTER:600:0:32 DS:ds1:COUNTER:600:0:32/ );
	$err = RRDs::error; if($err) { print "$id-rdy RRD Error: $err\n"; return -1; }
	RRDs::create( "$rrd-mem.rrd",	# actv/total
		qw/RRA:AVERAGE:0.5:1:800 RRA:AVERAGE:0.25:6:800 RRA:AVERAGE:0.25:24:800 RRA:AVERAGE:0.25:288:800 RRA:MAX:0.5:1:800 RRA:MAX:0.25:6:800 RRA:MAX:0.25:24:800 RRA:MAX:0.25:288:800/,
		qw/DS:ds0:GAUGE:600:0:1024000000000 DS:ds1:GAUGE:600:0:1024000000000/ );
	$err = RRDs::error; if($err) { print "$id-mem RRD Error: $err\n"; return -1; }
	RRDs::create( "$rrd-mem-ps.rrd", # pvt/shr
		qw/RRA:AVERAGE:0.5:1:800 RRA:AVERAGE:0.25:6:800 RRA:AVERAGE:0.25:24:800 RRA:AVERAGE:0.25:288:800 RRA:MAX:0.5:1:800 RRA:MAX:0.25:6:800 RRA:MAX:0.25:24:800 RRA:MAX:0.25:288:800/,
		qw/DS:ds0:GAUGE:600:0:1024000000000 DS:ds1:GAUGE:600:0:1024000000000/ );
	$err = RRDs::error; if($err) { print "RRD Error: $err\n"; return -1; }
	RRDs::create( "$rrd-mem-bs.rrd", # bal/swp
		qw/RRA:AVERAGE:0.5:1:800 RRA:AVERAGE:0.25:6:800 RRA:AVERAGE:0.25:24:800 RRA:AVERAGE:0.25:288:800 RRA:MAX:0.5:1:800 RRA:MAX:0.25:6:800 RRA:MAX:0.25:24:800 RRA:MAX:0.25:288:800/,
		qw/DS:ds0:GAUGE:600:0:1024000000000 DS:ds1:GAUGE:600:0:1024000000000/ );
	$err = RRDs::error; if($err) { print "RRD Error: $err\n"; return -1; }
	RRDs::create( "$rrd-rpc.rrd", # rpc
		qw/RRA:AVERAGE:0.5:1:800 RRA:AVERAGE:0.25:6:800 RRA:AVERAGE:0.25:24:800 RRA:AVERAGE:0.25:288:800 RRA:MAX:0.5:1:800 RRA:MAX:0.25:6:800 RRA:MAX:0.25:24:800 RRA:MAX:0.25:288:800/,
		qw/DS:ds0:COUNTER:600:0:1024000 DS:ds1:COUNTER:600:0:1024000/ );
	$err = RRDs::error; if($err) { print "RRD Error: $err\n"; return -1; }
	RRDs::create( "$rrd-swap.rrd", # rpc
		qw/RRA:AVERAGE:0.5:1:800 RRA:AVERAGE:0.25:6:800 RRA:AVERAGE:0.25:24:800 RRA:AVERAGE:0.25:288:800 RRA:MAX:0.5:1:800 RRA:MAX:0.25:6:800 RRA:MAX:0.25:24:800 RRA:MAX:0.25:288:800/,
		qw/DS:ds0:COUNTER:600:0:1024000 DS:ds1:COUNTER:600:0:1024000/ );
	$err = RRDs::error; if($err) { print "RRD Error: $err\n"; return -1; }
	# add entries to %cache
	$MAXID = $id;
	$cache{maxid} = $MAXID;
	$cache{"guest-$guest"} = $id;
	$cache{"$id-last"} = time;
	# return new ID
	return $id;
}
sub addserver($) {
	my($server) = $_[0];
	my($rrd,$err);
	# Prep from cache
	$srvdata{$server}{maxmem} = $cache{"srv-$server-maxmem"}
		if($cache{"srv-$server-maxmem"});
	$srvdata{$server}{maxcpu} = $cache{"srv-$server-maxcpu"}
		if($cache{"srv-$server-maxcpu"});
	# Create RRD for: cpu(sys,vhosts),cpu(ready),mem(reserved,console)
	#         swap(pi,po)
	$rrd = "$RRD/$FARM/srv-$server";
	if(! -f "$rrd-cpu.rrd" ) {
		print "Creating RRD files for $server\n" if($DEBUG);
		RRDs::create( "$rrd-cpu.rrd",
			qw/RRA:AVERAGE:0.5:1:800 RRA:AVERAGE:0.25:6:800 RRA:AVERAGE:0.25:24:800 RRA:AVERAGE:0.25:288:800 RRA:MAX:0.5:1:800 RRA:MAX:0.25:6:800 RRA:MAX:0.25:24:800 RRA:MAX:0.25:288:800/,
			qw/DS:ds0:COUNTER:600:0:32 DS:ds1:COUNTER:600:0:32/ );
		$err = RRDs::error; if($err) { print "RRD Error: $err\n"; }
		$cache{"srv-$server-last"} = time; # initial setup
	}
	if(! -f "$rrd-rdy.rrd" ) {
		RRDs::create( "$rrd-rdy.rrd",
			qw/RRA:AVERAGE:0.5:1:800 RRA:AVERAGE:0.25:6:800 RRA:AVERAGE:0.25:24:800 RRA:AVERAGE:0.25:288:800 RRA:MAX:0.5:1:800 RRA:MAX:0.25:6:800 RRA:MAX:0.25:24:800 RRA:MAX:0.25:288:800/,
			qw/DS:ds0:COUNTER:600:0:32 DS:ds1:COUNTER:600:0:32/ );
		$err = RRDs::error; if($err) { print "RRD Error: $err\n"; }
	}
	if(! -f "$rrd-mem.rrd" ) {
		RRDs::create( "$rrd-mem.rrd",	
			qw/RRA:AVERAGE:0.5:1:800 RRA:AVERAGE:0.25:6:800 RRA:AVERAGE:0.25:24:800 RRA:AVERAGE:0.25:288:800 RRA:MAX:0.5:1:800 RRA:MAX:0.25:6:800 RRA:MAX:0.25:24:800 RRA:MAX:0.25:288:800/,
			qw/DS:ds0:GAUGE:600:0:1024000000000 DS:ds1:GAUGE:600:0:1024000000000/ );
		$err = RRDs::error; if($err) { print "RRD Error: $err\n"; }
	}
	if(! -f "$rrd-swap.rrd" ) {
		RRDs::create( "$rrd-swap.rrd",	
			qw/RRA:AVERAGE:0.5:1:800 RRA:AVERAGE:0.25:6:800 RRA:AVERAGE:0.25:24:800 RRA:AVERAGE:0.25:288:800 RRA:MAX:0.5:1:800 RRA:MAX:0.25:6:800 RRA:MAX:0.25:24:800 RRA:MAX:0.25:288:800/,
			qw/DS:ds0:GAUGE:600:0:1024000 DS:ds1:GAUGE:600:0:1024000/ );
		$err = RRDs::error; if($err) { print "RRD Error: $err\n"; }
	}
}

#############################################################################
# Main subroutines
sub fetchdata($) {
	# collect guest and server info from SNMP daemon on this host
	my($server) = $_[0];
	my($snmp,$snmperr,$resp);
	my(%vmtoid) = ();
	my(%idtovm) = ();
	my($k,$vm,$id);
	my($guesthostname);
	my($now);

	%stats = ();
	print "Opening SNMP connection to $server\n" if($DEBUG);
	($snmp,$snmperr) = Net::SNMP->session( -hostname=>$server,
		-community=>$COMMUNITY, -timeout=>$TIMEOUT );
	if($snmperr) {
		print "SNMP Error: $snmperr\n";
		exit(1);
	}

	print "Fetching vmware-stats output...\n" if($DEBUG);
	$now = time;
	$resp = $snmp->get_table( -baseoid=>"$UCDOID.101"); # all stat data
	if(!$resp) {
		print "Error: Unable to retrieve vmware-stats data via SNMP from $server\n";
		return 1;
	}
	foreach my $oid ( keys %$resp ) {
		if( $resp->{$oid} =~ /^(\S+)=(.*)/) { $stats{$1}=$2; }
	}
	print "...done.\n" if($DEBUG);
	# %stats now holds all the info for this ESX server.

	# now we check to see if we've met this server before.
	# this does nothing if the rrd files already exist
	addserver($server);

	# We should now process it and add it into the %vmdata hash
	# if a guest is not recognised by getid(), call addguest()
	if( ! $stats{'has-names'} or !$stats{'version'}
		or $stats{'version'} < $MINVER ) {
		print "Error: vmware-stats agent is too old (need V$MINVER.\n";
		return 1;
	}
	foreach $k ( keys %stats ) {
		if( $k =~ /vhost-(\d+)-name/ ) {
			$vm = $1;
			$id = getid($stats{$k});
			$id = addguest($stats{$k}) if(!$id);
			if($id>-1) {
				$vmtoid{$vm} = $id;
				$idtovm{$id} = $vm;
			}
		}
	}
	foreach $id ( keys %idtovm ) {
		$vm = $idtovm{$id};
		$vmdata{$id}{state} = 1; # up
		$vmdata{$id}{'time'} = $stats{'time'};
		$vmdata{$id}{vmid} = $vm;
		$vmdata{$id}{server} = $server;
		$vmdata{$id}{rpc} = $stats{"vhost-$vm-rpc-count"};
		$vmdata{$id}{swapin} = $stats{"vhost-$vm-swap-in"};

		$cache{"$id-maxcpu"} = $vmdata{$id}{maxcpu} 
			= $stats{"vhost-$vm-cpu-max"};
		foreach ( qw/sys wait used ready/ ) {
			$vmdata{$id}{$_} = int($stats{"vhost-$vm-cpu-$_-count"})
				if(defined $stats{"vhost-$vm-cpu-$_-count"});
		}

		$cache{"$id-maxmem"} = $vmdata{$id}{maxmem} 
			= $stats{"vhost-$vm-mem-max"};
		foreach ( qw/shared private balloon swap active used/ ) {
			$vmdata{$id}{"mem-$_"} = int($stats{"vhost-$vm-mem-$_"})
				if(defined $stats{"vhost-$vm-mem-$_"});
		}

		$guesthostname = $stats{"vhost-$vm-name"};
		$vmdata{$id}{name} = $guesthostname;
		$guesthostname =~ s/\s+.*$//; $guesthostname = lc $guesthostname;
		if($guesthostname) {
			foreach ( glob("$MRTG/*/$guesthostname*.cfg") ) {
				if( -f ) { 
					print " Found MRTG config for $guesthostname\n" if($DEBUG);
					$vmdata{$id}{mrtgfile} = $_; 
					$vmdata{$id}{mrtgfile} =~ s/^$MRTG\///; 
					$vmdata{$id}{hostname} = $guesthostname;
					last; 
				}
			}
		}
		$vmdata{$id}{lastseen} = $cache{"$id-lastseen"} = $now;
	}
	# Now we do the per-server information
	$cache{"srv-$server-maxcpu"} 
		= $srvdata{$server}{maxcpu} = $stats{'sys-cpu-max'};
	$cache{"srv-$server-maxmem"} 
		= $srvdata{$server}{maxmem} = int($stats{'mem-avail'});
	$srvdata{$server}{'time'} = $stats{'time'};
	$srvdata{$server}{memfree} = int($stats{'mem-free'});
	$srvdata{$server}{sys} = int($stats{'sys-cpu-sys-count'});
	$srvdata{$server}{'wait'} = int($stats{'sys-cpu-wait-count'});
	$srvdata{$server}{used} = int($stats{'sys-cpu-used-count'});
	$srvdata{$server}{ready} = int($stats{'sys-cpu-ready-count'});
	$srvdata{$server}{swapin} = int($stats{"allvms-swap-in"});
	$srvdata{$server}{rpc} = int($stats{"allvms-rpc-count"})
		if( defined $stats{"allvms-rpc-count"} );

	$cache{"srv-$server-lastseen"} = $now;

	# return
	return 0;
}

########################################################################
# Update the RRD files from the data held in %vmdata and %vmware
sub updatehostrrd($) {
	my($server) = $_[0]; #  server name
	my($time);
	my($rrd,$err);
	$rrd = "$RRD/$FARM/srv-$server";
	$time = $srvdata{$server}{'time'};
	$time = time if(!$time);

	if(defined $cache{"srv-$server-last"}
		and $cache{"srv-$server-last"} >= $time ) {
		print "Skipping $server update because time goes backwards.\n"
			if($cache{"srv-$server-last"} > $time );
		print "Cache:".localtime($cache{"srv-$server-last"})
			."  Current:".localtime($time)."\n";
		return;
	}
	$cache{"srv-$server-last"} = $time;
	# Now update all the RRD files using the data from $srvdata{$server}
	# cpu - rdy - mem - swap
	print "Updating $server for ".localtime($time)."\n" if($DEBUG);
	RRDs::update("$rrd-cpu.rrd","--template","ds0:ds1",
		"$time:".$srvdata{$server}{used}.":".$srvdata{$server}{sys})	
		if(defined $srvdata{$server}{used});
	$err = RRDs::error; if($err) { print "cpu RRD Error: $err\n"; }
	RRDs::update("$rrd-rdy.rrd","--template","ds0:ds1",
		"$time:".$srvdata{$server}{ready}.":".$srvdata{$server}{ready})	
		if(defined $srvdata{$server}{ready});
	$err = RRDs::error; if($err) { print "rdy RRD Error: $err\n"; }
	RRDs::update("$rrd-mem.rrd","--template","ds0:ds1",
		"$time:".$srvdata{$server}{memfree}.":".$srvdata{$server}{maxmem})	
		if(defined $srvdata{$server}{memfree});
	$err = RRDs::error; if($err) { print "mem RRD Error: $err\n"; }
	RRDs::update("$rrd-swap.rrd","--template","ds0:ds1",
		"$time:".$srvdata{$server}{swapin}.":".$srvdata{$server}{swapin})	
		if(defined $srvdata{$server}{swapin});
	$err = RRDs::error; if($err) { print "swap RRD Error: $err\n"; }
	# Maybe do RPC here as well one day
}
sub updateguestrrd($) {
	my($id) = $_[0]; # our ID
	my($time);
	my($rrd,$err);
	$rrd = "$RRD/$FARM/$id";
	$time = $vmdata{$id}{'time'};

	$time = time if(!$time);
	if(defined $cache{"$id-last"} 
		and $cache{"$id-last"} >= $time ) {
		print "Skipping update of guest $id because time goes backwards.\n"
			if($cache{"$id-last"} > $time );
		return;
	}
	$cache{"$id-last"} = $time;

	# Now update all the RRD files using the data from $vmdata{$id}
	RRDs::update("$rrd-swap.rrd","--template","ds0:ds1",
		"$time:".$vmdata{$id}{swapin}.":".$vmdata{$id}{swapin})	
		if(defined $vmdata{$id}{swapin});
	$err = RRDs::error; if($err) { print "RRD Error: $id-swap: $err\n";}
	RRDs::update("$rrd-cpu.rrd","--template","ds0:ds1",
		"$time:".$vmdata{$id}{used}.":".$vmdata{$id}{sys})	
		if(defined $vmdata{$id}{used});
	$err = RRDs::error; if($err) { print "RRD Error: $id-cpu: $err\n"; }
	RRDs::update("$rrd-rdy.rrd","--template","ds0:ds1",
		"$time:".$vmdata{$id}{ready}.":".$vmdata{$id}{ready})	
		if(defined $vmdata{$id}{ready});
	$err = RRDs::error; if($err) { print "RRD Error: $id-rdy: $err\n"; }
	RRDs::update("$rrd-mem.rrd","--template","ds0:ds1",
		"$time:".$vmdata{$id}{"mem-used"}.":".$vmdata{$id}{"mem-active"})	
		if(defined $vmdata{$id}{"mem-used"});
	$err = RRDs::error; if($err) { print "RRD Error: $id-mem: $err\n"; }
	RRDs::update("$rrd-mem-ps.rrd","--template","ds0:ds1",
		"$time:".$vmdata{$id}{"mem-private"}.":".$vmdata{$id}{"mem-shared"})	
		if(defined $vmdata{$id}{"mem-private"});
	$err = RRDs::error; if($err) { print "RRD Error: $id-mem-ps: $err\n"; }
	RRDs::update("$rrd-mem-bs.rrd","--template","ds0:ds1",
		"$time:".$vmdata{$id}{"mem-balloon"}.":".$vmdata{$id}{"mem-swap"})	
		if(defined $vmdata{$id}{"mem-balloon"});
	$err = RRDs::error; if($err) { print "RRD Error: $id-mem-bs: $err\n"; }
}
########################################################################
# Scripts to generate conf files

sub cfgheader {
	my($name) = $_[0];
	print CFG "DO NOT RUN IN MRTG: for frontend use only\n";
	print CFG "#\n# Config for: $name\n#\n#This file is autogenerated.  Do not change it!\n";
	print CFG "WorkDir: $RRD/$FARM\n";
	print CFG "LogFormat: rrdtool\n";
	print CFG "Options[_]: growright\n";
    print CFG "routers.cgi*RoutingTable: no\n";
    print CFG "routers.cgi*Icon: vmware-sm.gif\n";
    print CFG "routers.cgi*NoCache: yes\n";
    print CFG "EnableIPv6: no\n";
	print CFG "# End of standard header\n\n";
}
sub writeserverdetail($$) {
	my($server,$quiet) = @_; # quiet!=0 if this is for the server not the farm

	# server memory stats

	# XXX active

	# server CPU stats
	if( $srvdata{$server}{maxcpu} ) {
		print CFG "# CPU: percentage of CPU in use\n";
		print CFG "Target[srv-$server-cpu]: cpu\@$server\n";
		print CFG "Title[srv-$server-cpu]: CPU usage on $server\["
			.$srvdata{$server}{maxcpu}."\]\n";
		print CFG "routers.cgi*ShortName[srv-$server-cpu]: CPU Use\n";
		print CFG "MaxBytes[srv-$server-cpu]: ".$srvdata{$server}{maxcpu}."\n";
		print CFG "Factor[srv-$server-cpu]: ".(100/$srvdata{$server}{maxcpu})."\n";
		print CFG "Options[srv-$server-cpu]: growright \n";
		print CFG "Legend1[srv-$server-cpu]: CPU usage\n";
		print CFG "Legend2[srv-$server-cpu]: CPU sys\n";
		print CFG "Legend3[srv-$server-cpu]: Max CPU usage\n";
		print CFG "Legend4[srv-$server-cpu]: Max CPU sys\n";
		print CFG "YLegend[srv-$server-cpu]: percent\n";
		print CFG "ShortLegend[srv-$server-cpu]: %\n";
		print CFG "LegendI[srv-$server-cpu]: used:\n";
		print CFG "LegendO[srv-$server-cpu]: sys :\n";
		print CFG "routers.cgi*GraphStyle[srv-$server-cpu]: stack\n";
		if($quiet) {
			print CFG "routers.cgi*InMenu[srv-$server-cpu]: no \n";
			print CFG "routers.cgi*InSummary[srv-$server-cpu]: no \n";
			print CFG "routers.cgi*InCompact[srv-$server-cpu]: no \n";
			print CFG "routers.cgi*Summary[srv-$server-cpu]: svr-cpu-summary\n";
			print CFG "routers.cgi*Graph[srv-$server-cpu]: svr-cpu\n";
		} else {
			print CFG "routers.cgi*InMenu[srv-$server-cpu]: yes\n";
			print CFG "routers.cgi*InSummary[srv-$server-cpu]: yes\n";
			print CFG "routers.cgi*InCompact[srv-$server-cpu]: yes\n";
		}
		print CFG "routers.cgi*InOut[srv-$server-cpu]: no\n";
		print CFG "routers.cgi*Options[srv-$server-cpu]: nopercent scaled fixunit nototal\n";
		print CFG "routers.cgi*Mode[srv-$server-cpu]: generic\n";
		#  ready
		print CFG "Target[srv-$server-rdy]: ready\@$server\n";
		print CFG "Title[srv-$server-rdy]: CPU ready time on $server\[".$srvdata{$server}{maxcpu}."\]\n";
		print CFG "routers.cgi*ShortName[srv-$server-rdy]: CPU ready\n";
		print CFG "MaxBytes[srv-$server-rdy]: ".$srvdata{$server}{maxcpu}."\n";
		print CFG "Factor[srv-$server-rdy]: ".(100/$srvdata{$server}{maxcpu})."\n";
		print CFG "Options[srv-$server-rdy]: growright noo\n";
		print CFG "Legend1[srv-$server-rdy]: CPU ready\n";
		print CFG "Legend3[srv-$server-rdy]: Max CPU ready\n";
		print CFG "YLegend[srv-$server-rdy]: percent\n";
		print CFG "ShortLegend[srv-$server-rdy]: %\n";
		print CFG "LegendI[srv-$server-rdy]: ready:\n";
		if($quiet) {
			print CFG "routers.cgi*InMenu[srv-$server-rdy]: no \n";
			print CFG "routers.cgi*InSummary[srv-$server-rdy]: no \n";
			print CFG "routers.cgi*InCompact[srv-$server-rdy]: no \n";
			print CFG "routers.cgi*Summary[srv-$server-rdy]: svr-ready-summary\n";
			print CFG "routers.cgi*Graph[srv-$server-rdy]: svr-ready\n";
		} else {
			print CFG "routers.cgi*InMenu[srv-$server-rdy]: yes\n";
			print CFG "routers.cgi*InSummary[srv-$server-rdy]: yes\n";
			print CFG "routers.cgi*InCompact[srv-$server-rdy]: yes\n";
		}
		print CFG "routers.cgi*InOut[srv-$server-rdy]: no\n";
		print CFG "routers.cgi*Options[srv-$server-rdy]: nopercent scaled fixunit nototal\n";
		print CFG "routers.cgi*Mode[srv-$server-rdy]: generic\n";
	}

	# Server swap stats
	# XXX

	# Server RPC
	# XXX
}
sub writeguestdetail($$) {
	my($guest,$quiet) = @_;
	my($name,$server);
	$name = $vmdata{$guest}{name}; $name =~ s/\s+.*$//;
	$server = $vmdata{$guest}{server};

	# Now create Target for each RRD
	if( $vmdata{$guest}{maxcpu} ) {
		print CFG "# CPU: percentage of allocated CPU in use\n";
		print CFG "Target[$guest-cpu]: $name\@$server\n";
		print CFG "Title[$guest-cpu]: Allocated CPU usage on $name\[".$vmdata{$guest}{maxcpu}."\]\n";
		print CFG "routers.cgi*ShortName[$guest-cpu]: CPU Use\n";
		print CFG "MaxBytes[$guest-cpu]: ".$vmdata{$guest}{maxcpu}."\n";
		print CFG "Factor[$guest-cpu]: ".(100/$vmdata{$guest}{maxcpu})."\n";
		print CFG "Options[$guest-cpu]: growright \n";
		print CFG "Legend1[$guest-cpu]: CPU usage\n";
		print CFG "Legend2[$guest-cpu]: CPU sys\n";
		print CFG "Legend3[$guest-cpu]: Max CPU usage\n";
		print CFG "Legend4[$guest-cpu]: Max CPU sys\n";
		print CFG "YLegend[$guest-cpu]: percent\n";
		print CFG "ShortLegend[$guest-cpu]: %\n";
		print CFG "LegendI[$guest-cpu]: used:\n";
		print CFG "LegendO[$guest-cpu]: sys :\n";
		print CFG "routers.cgi*GraphStyle[$guest-cpu]: stack\n";
		print CFG "routers.cgi*InOut[$guest-cpu]: no\n";
		if($quiet){
			print CFG "routers.cgi*InMenu[$guest-cpu]: no \n";
			print CFG "routers.cgi*InSummary[$guest-cpu]: no \n";
			print CFG "routers.cgi*InCompact[$guest-cpu]: no \n";
			print CFG "routers.cgi*Graph[$guest-cpu]: guest-cpu noo\n";
		} else {
			print CFG "routers.cgi*InMenu[$guest-cpu]: yes\n";
			print CFG "routers.cgi*InSummary[$guest-cpu]: yes\n";
			print CFG "routers.cgi*InCompact[$guest-cpu]: yes\n";
		}
		print CFG "routers.cgi*Options[$guest-cpu]: nopercent scaled fixunit nototal\n";
		print CFG "routers.cgi*Mode[$guest-cpu]: generic\n";

		print CFG "Target[$guest-rdy]: $name\@$server\n";
		print CFG "Title[$guest-rdy]: Allocated CPU ready time on $name\[".$vmdata{$guest}{maxcpu}."\]\n";
		print CFG "routers.cgi*ShortName[$guest-rdy]: CPU ready\n";
		print CFG "MaxBytes[$guest-rdy]: ".$vmdata{$guest}{maxcpu}."\n";
		print CFG "Factor[$guest-rdy]: ".(100/$vmdata{$guest}{maxcpu})."\n";
		print CFG "Options[$guest-rdy]: growright noo\n";
		print CFG "Legend1[$guest-rdy]: CPU ready\n";
		print CFG "Legend3[$guest-rdy]: Max CPU ready\n";
		print CFG "YLegend[$guest-rdy]: percent\n";
		print CFG "ShortLegend[$guest-rdy]: %\n";
		print CFG "LegendI[$guest-rdy]: ready:\n";
		print CFG "routers.cgi*InOut[$guest-rdy]: no\n";
		if($quiet){
			print CFG "routers.cgi*InMenu[$guest-rdy]: no \n";
			print CFG "routers.cgi*InSummary[$guest-rdy]: no \n";
			print CFG "routers.cgi*InCompact[$guest-rdy]: no \n";
			print CFG "routers.cgi*Graph[$guest-rdy]: guest-ready noo \n";
		}else{
			print CFG "routers.cgi*InMenu[$guest-rdy]: yes\n";
			print CFG "routers.cgi*InSummary[$guest-rdy]: yes\n";
			print CFG "routers.cgi*InCompact[$guest-rdy]: yes\n";
		}
		print CFG "routers.cgi*Options[$guest-rdy]: nopercent scaled fixunit nototal\n";
		print CFG "routers.cgi*Mode[$guest-rdy]: generic\n";
	} else {
		print CFG "# Cannot add CPU graphs as no max-cpu for vhost\n";
	}
	
	# Memory definitions

	if( $vmdata{$guest}{maxmem} ) {
		print CFG "# Memory usage (current: "
			.$vmdata{$guest}{"mem-used"}."/".$vmdata{$guest}{"mem-active"} 
			.")\n"
			if(defined $vmdata{$guest}{"mem-active"} );
		print CFG "Target[$guest-mem]: $name\@$server\n";
		print CFG "Title[$guest-mem]: Memory in use on $name\n";
		print CFG "routers.cgi*ShortName[$guest-mem]: Memory\n";
		print CFG "MaxBytes[$guest-mem]: ".$vmdata{$guest}{maxmem}."\n";
		print CFG "Options[$guest-mem]: growright gauge\n";
		print CFG "Legend1[$guest-mem]: Memory usage\n";
		print CFG "Legend2[$guest-mem]: Memory active\n";
		print CFG "Legend3[$guest-mem]: Max memory usage\n";
		print CFG "Legend4[$guest-mem]: Max memory active\n";
		print CFG "YLegend[$guest-mem]: Bytes\n";
		print CFG "ShortLegend[$guest-mem]: B\n";
		print CFG "LegendI[$guest-mem]: used  :\n";
		print CFG "LegendO[$guest-mem]: active:\n";
		print CFG "routers.cgi*InOut[$guest-mem]: no\n";
		if($quiet) {
			print CFG "routers.cgi*InMenu[$guest-mem]: no \n";
			print CFG "routers.cgi*InSummary[$guest-mem]: no \n";
			print CFG "routers.cgi*InCompact[$guest-mem]: no \n";
			print CFG "routers.cgi*Graph[$guest-mem]: guest-mem noo \n";
		} else {
			print CFG "routers.cgi*InMenu[$guest-mem]: yes\n";
			print CFG "routers.cgi*InSummary[$guest-mem]: yes\n";
			print CFG "routers.cgi*InCompact[$guest-mem]: yes\n";
		}
		print CFG "routers.cgi*Options[$guest-mem]: scaled bytes \n";

		if(!$quiet) {
		print CFG "# Memory breakdown\n";
		print CFG "# (current: "
			.$vmdata{$guest}{"mem-private"}."/"
			.$vmdata{$guest}{"mem-shared"}."/"
			.$vmdata{$guest}{"mem-balloon"}."/"
			.$vmdata{$guest}{"mem-swap"}.")\n"
			if(defined $vmdata{$guest}{"mem-private"});
		print CFG "Target[$guest-mem-ps]: $name(ps)\@$server\n";
		print CFG "Title[$guest-mem-ps]: Memory in use on $name (private, shared)\n";
		print CFG "routers.cgi*ShortName[$guest-mem-ps]: $name Memory\n";
		print CFG "MaxBytes[$guest-mem-ps]: ".$vmdata{$guest}{maxmem}."\n";
		print CFG "Options[$guest-mem-ps]: growright gauge\n";
		print CFG "Legend1[$guest-mem-ps]: Private memory\n";
		print CFG "Legend2[$guest-mem-ps]: Shared memory\n";
		print CFG "Legend3[$guest-mem-ps]: Max private\n";
		print CFG "Legend4[$guest-mem-ps]: Max shared\n";
		print CFG "YLegend[$guest-mem-ps]: Bytes\n";
		print CFG "ShortLegend[$guest-mem-ps]: B\n";
		print CFG "LegendI[$guest-mem-ps]: private:\n";
		print CFG "LegendO[$guest-mem-ps]: shared :\n";
		print CFG "routers.cgi*InMenu[$guest-mem-ps]: no\n";
		print CFG "routers.cgi*InOut[$guest-mem-ps]: no\n";
		print CFG "routers.cgi*InSummary[$guest-mem-ps]: no\n";
		print CFG "routers.cgi*InCompact[$guest-mem-ps]: yes\n";
		print CFG "routers.cgi*Options[$guest-mem-ps]: scaled bytes \n";
		print CFG "routers.cgi*Graph[$guest-mem-ps]: memsplit \n";

		print CFG "Target[$guest-mem-bs]: $name(bs)\@$server\n";
		print CFG "Title[$guest-mem-bs]: Memory in use on $name (swapped, balloon)\n";
		print CFG "routers.cgi*ShortName[$guest-mem-bs]: $name Memory\n";
		print CFG "MaxBytes[$guest-mem-bs]: ".$vmdata{$guest}{maxmem}."\n";
		print CFG "Options[$guest-mem-bs]: growright gauge\n";
		print CFG "Legend2[$guest-mem-bs]: Swapped memory\n";
		print CFG "Legend1[$guest-mem-bs]: Balloon memory\n";
		print CFG "Legend4[$guest-mem-bs]: Max swapped\n";
		print CFG "Legend3[$guest-mem-bs]: Max balloon\n";
		print CFG "YLegend[$guest-mem-bs]: Bytes\n";
		print CFG "ShortLegend[$guest-mem-bs]: b\n";
		print CFG "LegendO[$guest-mem-bs]: swapped:\n";
		print CFG "LegendI[$guest-mem-bs]: balloon:\n";
		print CFG "routers.cgi*InMenu[$guest-mem-bs]: no\n";
		print CFG "routers.cgi*InOut[$guest-mem-bs]: no\n";
		print CFG "routers.cgi*InSummary[$guest-mem-bs]: no\n";
		print CFG "routers.cgi*InCompact[$guest-mem-bs]: yes\n";
		print CFG "routers.cgi*Options[$guest-mem-bs]: scaled bytes \n";
		print CFG "routers.cgi*Graph[$guest-mem-bs]: memsplit \n";

		print CFG "routers.cgi*InMenu[memsplit]: yes\n";
		print CFG "routers.cgi*GraphStyle[memsplit]: stack\n";
		print CFG "routers.cgi*Title[memsplit]: Memory split for $name\n";
		print CFG "routers.cgi*ShortName[memsplit]: Memory split\n";
		print CFG "routers.cgi*Icon[memsplit]: chip-sm.gif\n";
		print CFG "routers.cgi*Colours[memsplit]: #0000ff #00ff00 #cccc00 #ff0000\n";
		} # not quiet mode
	} else {
		print CFG "# Cannot give memory as no max available\n";
	}

	# XXX
	# rpc
	# swap

}
sub writeguestcfg($) {
	# Generate the MRTG .conf file for this guest (key to %vmdata)
	# This will be relatively small with only a few graphs
	# plus possibly a link to the real server info.
	my($guest) = $_[0]; # our ID
	my($cfgfile) = "$MRTG/$SUBDIR/$FARM/guests/$guest.conf";
	my($name,$server);

	if(!$guest or !defined $vmdata{$guest}) { print "Bad guest ID\n"; return; }
	$vmdata{$guest}{server} = "NONE" if(! $vmdata{$guest}{server} );
	$vmdata{$guest}{name} = "Unknown" if(! $vmdata{$guest}{name} );

	$name = $vmdata{$guest}{name}; $name =~ s/\s+.*$//;
	$server = $vmdata{$guest}{server};

	open CFG,">$cfgfile" or do { print "$cfgfile: $!\n"; return; };
	print CFG "# This guest last seen alive "
		.localtime($vmdata{$guest}{lastseen})."\n" 
			if($vmdata{$guest}{lastseen});
	cfgheader($vmdata{$guest}{name}." ($guest) on "
		.$vmdata{$guest}{server}." in ESX farm $FARM");
	if($vmdata{$guest}{hostname}) {
	    print CFG "routers.cgi*ShortDesc: ".$vmdata{$guest}{hostname}."\n";
	} else {
	    print CFG "routers.cgi*ShortDesc: $name\n";
	}
    print CFG "routers.cgi*Description: ".$vmdata{$guest}{name}."\n";
    print CFG "routers.cgi*DefaultTarget: summary\n";
	print CFG "routers.cgi*Link: \"Host Stats\" ".$vmdata{$guest}{mrtgfile}." _summary server-sm.gif\n"
		if($vmdata{$guest}{mrtgfile});
	print CFG "routers.cgi*Link: \"ESX Server\" $SUBDIR/$FARM/$server.conf _summary esx-sm.gif\n"
		if($server);
	print CFG "routers.cgi*Link: \"Farm\" $SUBDIR/$FARM.conf _none esx-sm.gif\n";
	print CFG "PageFoot[_]: <I><SMALL>Last seen alive at: ".localtime($vmdata{$guest}{lastseen})."</SMALL></I>\n"
		if($vmdata{$guest}{lastseen});
	print CFG "\n";

	writeguestdetail($guest,0);

	close CFG;
}
sub writehostcfg($) {
	# Generate the MRTG .conf file for this ESX server using %srvdata hash
	# This will be relatively small with only server summay info
	my($server) = $_[0];
	my($cfgfile) = "$MRTG/$SUBDIR/$FARM/$server.conf";

	open CFG,">$cfgfile" or do { print "$cfgfile: $!\n"; return; };
	cfgheader("$server in ESX farm $FARM");
    print CFG "routers.cgi*ShortDesc: $server\n";
    print CFG "routers.cgi*DefaultTarget: info\n";
    print CFG "routers.cgi*Description: ESX Server $server\n";
    print CFG "routers.cgi*Extension: Management https://$server$DOMAIN/ cog-sm.gif _new noopts\n";
	print CFG "routers.cgi*Link: \"Farm\" $SUBDIR/$FARM.conf _none esx-sm.gif\n";
	print CFG "\n";

	writeserverdetail($server,0);

	close CFG;
}
sub writefarmcfg {
	# Generate the MRTG .conf file for this farm's summary
	# totals for servers and for all guests
	# This is the big one.  Include all the guests' graphs in summaries,
	# plus all the servers' graph in summaries.
	my($cfgfile) = "$MRTG/$SUBDIR/$FARM.conf";

	open CFG,">$cfgfile" or do { print "$cfgfile: $!\n"; return; };
	cfgheader("ESX farm $FARM");
    print CFG "routers.cgi*ShortDesc: $FARM\n";
    print CFG "routers.cgi*Description: ESX Farm $FARM\n";
    print CFG "routers.cgi*DefaultTarget: info\n";
	print CFG "\n";
	foreach ( keys %vmdata ) {
		print CFG "# Guest $_: ".$vmdata{$_}{name}."\n";
		# Add the info for ALL the guest here
		writeguestdetail($_,1);

	}
	# Add the details for the userdefined summaries of the guests
	print CFG "routers.cgi*Desc[guest-cpu]: All guests CPU usage\n";
	print CFG "routers.cgi*ShortDesc[guest-cpu]: Guest CPU\n";
	print CFG "routers.cgi*Desc[guest-ready]: All guests CPU ready time\n";
	print CFG "routers.cgi*ShortDesc[guest-ready]: Guest Ready\n";
	print CFG "routers.cgi*Desc[guest-mem]: All guests memory usage\n";
	print CFG "routers.cgi*ShortDesc[guest-mem]: Guest Memory\n";

	foreach ( keys %srvdata ) {
		print CFG "# ESX Server $_: $_\n";
		# Add the info for the server here
		writeserverdetail($_,1);

	}
	# Add the details for the userdefined summaries of the servers
	print CFG "routers.cgi*Desc[svr-cpu]: All servers CPU usage\n";
	print CFG "routers.cgi*ShortDesc[svr-cpu]: Svr CPU Usage\n";
	print CFG "routers.cgi*Desc[svr-ready]: All servers CPU ready time\n";
	print CFG "routers.cgi*ShortDesc[svr-ready]: Svr Ready time\n";
	print CFG "routers.cgi*Desc[svr-mem]: All servers memory usage\n";
	print CFG "routers.cgi*ShortDesc[svr-mem]: Svr Memory\n";

	close CFG;
}
sub clearcfg {
	# Delete ALL the generated .conf files
	# But NOT the .rrd files or cache file
	print "Deleting old conf files\n" if($DEBUG);

	# Farm config file
	unlink "$MRTG/$SUBDIR/$FARM.conf" if(-f "$MRTG/$SUBDIR/$FARM.conf");

	# Servers
	foreach my $svr ( glob("$MRTG/$SUBDIR/$FARM/*.conf") ) { unlink $svr; }

	# Guests
	foreach my $guest ( glob("$MRTG/$SUBDIR/$FARM/guests/*.conf") ) { unlink $guest; }
}
#############################################################################
# Main code

getopts('gC:D:B:hdt:F:R:x');
if($opt_h or !@ARGV) { dohelp(); exit 0; }
$DEBUG = $opt_d if($opt_d);
$COMMUNITY = $opt_C if($opt_C);
$FARM = $opt_F if($opt_F);
$RRD = $opt_R if($opt_R and -d $opt_R);
$MRTG = $opt_D if($opt_D and -d $opt_D);
$SUBDIR = $opt_B if($opt_B);
$TIMEOUT = $opt_t if($opt_t);
if(! -d $MRTG) {
	print "MRTG directory $MRTG does not exist!\n";
	exit 1;
}
mkdir "$MRTG/$SUBDIR" if(!-d "$MRTG/$SUBDIR");
mkdir "$MRTG/$SUBDIR/$FARM" if(!-d "$MRTG/$SUBDIR/$FARM");
mkdir "$MRTG/$SUBDIR/$FARM/guests" if(!-d "$MRTG/$SUBDIR/$FARM/guests");
mkdir "$RRD" if(!-d "$RRD");
mkdir "$RRD/$FARM" if(!-d "$RRD/$FARM");

print "Processing farm $FARM\n";

%vmdata = ();  # no vmdata yet
%srvdata = (); # no server data yet either
print "Reading cache...\n" if($DEBUG);
readcache;
clearcfg if($opt_x);  # Maybe this should not be done? It will confuse routers2
print "Updating server files...\n" if($DEBUG);
foreach $HOST ( @ARGV ) {
	next if(fetchdata($HOST));
	writehostcfg($HOST);
	updatehostrrd($HOST);
}
print "Updating guest files...\n" if($DEBUG);
foreach $GUEST ( keys %vmdata ) { # keyed by ID
	writeguestcfg($GUEST) unless($opt_g); 
	updateguestrrd($GUEST); 
}
print "Updating farm files...\n" if($DEBUG);
writefarmcfg;
print "Writing cache...\n" if($DEBUG);
writecache;
exit 0;
