/* check_vcpu
vim:ts=4
	Nagios plugin for Linux guest running under VMWARE ESX3
	Steve Shipway, University of Auckland
 
	Incorporates some example code as shipped with SDK

	Requires VMware Tools and SDK
*/

#define VERSION "1.0"

/* Includes as given */
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <math.h>
#ifdef _WIN32
#	include <windows.h>
#else
#	include <unistd.h>
#	include <dlfcn.h>
#endif
#include "vmGuestLib.h"
#ifdef WIN32
#define SLEEP(x) Sleep(x * 1000)
#define TMPFILE "C:\\vmware.txt"
#else
#define SLEEP(x) sleep(x)
#define TMPFILE "/tmp/vmware-data"
#endif

/* for options */
extern char *optarg;
extern int optind, opterr, optopt;

/* Variables for the monitored data */
double percentage;
double warncpu = 70.0;
double critcpu = 90.0;
uint64 oldused = 0;
uint64 oldelapsed = 0;


static Bool done = FALSE;
/* Functions to dynamically load from the GuestLib library. */
char const * (*GuestLib_GetErrorText)(VMGuestLibError);
VMGuestLibError (*GuestLib_OpenHandle)(VMGuestLibHandle*);
VMGuestLibError (*GuestLib_CloseHandle)(VMGuestLibHandle);
VMGuestLibError (*GuestLib_UpdateInfo)(VMGuestLibHandle handle);
VMGuestLibError (*GuestLib_GetSessionId)(VMGuestLibHandle handle,
				                             VMSessionId *id);
VMGuestLibError (*GuestLib_GetCpuShares)(VMGuestLibHandle handle, uint32 *cpuShares);
VMGuestLibError (*GuestLib_GetCpuUsedMs)(VMGuestLibHandle handle, uint64 *cpuUsedMs);
VMGuestLibError (*GuestLib_GetHostProcessorSpeed)(VMGuestLibHandle handle, uint32 *mhz);
VMGuestLibError (*GuestLib_GetElapsedMs)(VMGuestLibHandle handle, uint64 *elapsedMs);

/*
 * Handle for use with shared library.
 */

#ifdef WIN32
HMODULE dlHandle = NULL;
#else
void *dlHandle = NULL;
#endif

/*
 * GuestLib handle.
 */
VMGuestLibHandle glHandle;


/*
 * Macro to load a single GuestLib function from the shared library.
 */

#ifdef _WIN32
#define LOAD_ONE_FUNC(funcname)				                          \
	do {			                                                     \
		(FARPROC)funcname = GetProcAddress(dlHandle, "VM" #funcname);  \
		if (funcname == NULL) {		                                  \
			error = GetLastError();	                                  \
			printf("Failed to load \'%s\': %d\n",	                    \
				    #funcname, error);                                   \
			return FALSE;	                                            \
		}		                                                        \
	} while (0)

#else

#define LOAD_ONE_FUNC(funcname)				               \
	do {			                                          \
		funcname = dlsym(dlHandle, "VM" #funcname);		   \
		if ((dlErrStr = dlerror()) != NULL) {		         \
			printf("Failed to load \'%s\': \'%s\'\n",	     \
				    #funcname, dlErrStr);                     \
			return FALSE;	                                 \
		}		                                             \
	} while (0)

#endif

void do_version() {
	printf("check_vcpu version %s\n",VERSION);
	exit(0);
}
void do_help() {
	printf("check_vcpu version %s\n",VERSION);
	printf("Usage: check_vcpu [-w warnpercent][-c critpercent]\n");
	printf("Nagios plugin for NRPE.  Requires ESX3 vmware-tools.\n");
	printf("Measures percentage of maximum allocated CPU being used by virtual guest OS\n");
	printf("S Shipway, University of Auckland, 2007\n");
	printf("Incorporates code from vmware-tools SDK, (c)VMWare\n");
	printf("All rights of VMWare to the SDK components required for compilation of this\nplugin are acknowledged.\n");
}
/* Load and save previous CPU polls for calculating CPU percentages */
void writedata(uint64 cpuUsedMs,uint64 elapsedMs) {
	FILE *fp;
	if( !(fp = fopen( TMPFILE,"w" )) ) { return; }
/*	fprintf(fp,"%"FMT64"u\n%"FMT64"u\n",cpuUsedMs,elapsedMs); */
	fwrite(&cpuUsedMs,sizeof(cpuUsedMs),1,fp);
	fwrite(&elapsedMs,sizeof(elapsedMs),1,fp);
	fclose(fp);
}
Bool readdata(void) {
	FILE *fp;
	if( !(fp = fopen( TMPFILE,"r" )) ) { return FALSE; }
	fread(&oldused   ,sizeof(oldused),1,fp);
	fread(&oldelapsed,sizeof(oldelapsed),1,fp);
	fclose(fp);
	return TRUE;
}

/*
 *-----------------------------------------------------------------------------
 *
 * LoadFunctions --
 *
 *		Load the functions from the shared library.
 *
 * Results:
 *		TRUE on success
 *		FALSE on failure
 *
 * Side effects:
 *		None
 *
 *-----------------------------------------------------------------------------
 */

Bool LoadFunctions(void) {
	/*
	 * First, try to load the shared library.
	 */
#ifdef _WIN32
	DWORD error;

	dlHandle = LoadLibrary("vmGuestLib.dll");
	if (!dlHandle) {
		error = GetLastError();
		printf("LoadLibrary failed: %d\n", error);
		return FALSE;
	}
#else
	char const *dlErrStr;

	dlHandle = dlopen("libvmGuestLib.so", RTLD_NOW);
	if (!dlHandle) {
		dlErrStr = dlerror();
		printf("dlopen failed: \'%s\'\n", dlErrStr);
		return FALSE;
	}
#endif

	/* Load all the individual library functions. */
	LOAD_ONE_FUNC(GuestLib_GetErrorText);
	LOAD_ONE_FUNC(GuestLib_OpenHandle);
	LOAD_ONE_FUNC(GuestLib_CloseHandle);
	LOAD_ONE_FUNC(GuestLib_UpdateInfo);
	LOAD_ONE_FUNC(GuestLib_GetSessionId);
	LOAD_ONE_FUNC(GuestLib_GetCpuShares);
	LOAD_ONE_FUNC(GuestLib_GetCpuUsedMs);
	LOAD_ONE_FUNC(GuestLib_GetHostProcessorSpeed);
	LOAD_ONE_FUNC(GuestLib_GetElapsedMs);

	return TRUE;
}


/* Actually obtain the VMWare ESX statistics */
Bool getdata(void) {
	VMGuestLibError glError;
	Bool success = TRUE;
	uint32 cpuShares = 0;
	uint64 cpuUsedMs = 0;
	uint32 hostMHz = 0;
	uint64 elapsedMs = 0;
	VMSessionId sessionId = 0;
	char resourcePoolPath[512];
	size_t poolBufSize;

	/* Try to load the library. */
	glError = GuestLib_OpenHandle(&glHandle);
	if (glError != VMGUESTLIB_ERROR_SUCCESS) {
		printf("UNKNOWN: Is vmware-tools installed correctly?  OpenHandle failed: %s\n", GuestLib_GetErrorText(glError));
	exit(3);
	}

	/* Attempt to retrieve info from the host. */
		VMSessionId tmpSession;

		glError = GuestLib_UpdateInfo(glHandle);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("UNKNOWN: Is vmware-tools working correctly?  UpdateInfo failed: %s\n", GuestLib_GetErrorText(glError));
			exit(3);
		}
		/* Retrieve and check the session ID */
		glError = GuestLib_GetSessionId(glHandle, &tmpSession);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("UNKNOWN: Failed to get session ID: %s\n", GuestLib_GetErrorText(glError));	
			exit(3);
		}

		if (tmpSession == 0) {
			printf("UNKNOWN: Got zero sessionId from GuestLib.  Do you have the latest version of vmware-tools?\n");
			exit(3);
		}
		sessionId = tmpSession;
		/* Retrieve all the stats. */
		glError = GuestLib_GetCpuShares(glHandle, &cpuShares);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("Failed to get cpu shares: %s\n", GuestLib_GetErrorText(glError));
			exit(3); }
		glError = GuestLib_GetCpuUsedMs(glHandle, &cpuUsedMs);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("Failed to get used ms: %s\n", GuestLib_GetErrorText(glError));
			exit(3); }
		glError = GuestLib_GetHostProcessorSpeed(glHandle, &hostMHz);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("Failed to get host proc speed: %s\n", GuestLib_GetErrorText(glError));
			exit(3); }
		glError = GuestLib_GetElapsedMs(glHandle, &elapsedMs);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("Failed to get elapsed ms: %s\n", GuestLib_GetErrorText(glError));
			exit(3); }

		/* Now we have the data, we can report on it.
			We could report on memory or on CPU.  Default to CPU. */

		/* Now, our 100% limit is cpuLimitMHz/hostMHz ms per ms. 
			So to find out our % usage, we do delta used/(delta time x limit)*/

#ifdef DEBUG
	if(!readdata()) {
		printf("Failed to read in data.\n");   
	} else {
		printf("Old Used    : %ums\n",(oldused));
		printf("Old Elapsed : %ums\n",(oldelapsed)); 
	}
#else
	readdata();
#endif
	writedata(cpuUsedMs,elapsedMs);
#ifdef DEBUG
	printf("Used time   : %ums\n",(cpuUsedMs-oldused));
	printf("Elapsed time: %ums\n",(elapsedMs-oldelapsed));
	printf("CPU available %u\n",cpuShares);
	printf("CPU host    : %u\n",hostMHz);
	printf("CPU factor  : %.2f\n",((double)cpuShares/(double)hostMHz));
	printf("Used/ms     : %.4f\n",((double)(cpuUsedMs-oldused)/(double)(elapsedMs-oldelapsed)));

#endif
	percentage = (100.0*(double)(cpuUsedMs-oldused)/(double)(elapsedMs-oldelapsed))/((double)cpuShares/(double)hostMHz);

	percentage = floor(100.0*percentage)/100.0; 
#ifdef DEBUG
	printf("Percentage  : %.4f\n",percentage); 
#endif


	glError = GuestLib_CloseHandle(glHandle);
#ifdef DEBUG
	if (glError != VMGUESTLIB_ERROR_SUCCESS) {
		printf("UNKNOWN: Failed to CloseHandle: %s\n", GuestLib_GetErrorText(glError));
		exit(3);
	}
#endif
	return TRUE;
}

int main(int argc, char *argv[]) {
	int c;
    while( (c=getopt(argc,argv,"hvw:c:"))>0 ) {
		switch(c) {
			case 'h': do_help(); exit(0); break;
			case 'v': do_version(); exit(0); break;
			case 'w': warncpu = atof(optarg); break;
			case 'c': critcpu = atof(optarg); break;
			default: do_help; exit(0);
		}
	}
	/* Try to load the library. */
	if (!LoadFunctions()) {
		printf("UNKNOWN: Failed to load vmware-tools library.  Are you sure that this is host is an ESX3 Guest and vmware-tools is installed?\n");
		exit(3);
	}

	/* extract the performance metrics from vmware-tools */
	if (!getdata()) {
		printf("UNKNOWN: Retrieving data from ESX failed\n");
		exit(3);
	}

#ifdef _WIN32
	if (!FreeLibrary(dlHandle)) {
		DWORD error = GetLastError();
		printf("UNKNOWN: Failed to FreeLibrary: %d\n", error);
		exit(3);
	}
#else
	if (dlclose(dlHandle)) {
		printf("UNKNOWN: dlclose failed for some reason\n");
		exit(3);
	}
#endif

	if((percentage>100.0)||(percentage<0.0)) {
		printf("UNKNOWN: CPU Usage cannot be determined.\n");
		exit(3); }

	/* now we have the data!  We can do thresholding, and all sorts of 
	   other groovy stuff */
	if( percentage >= critcpu ) {
		printf( "CRIT: CPU usage is %.2f%% of allocated maximum\n",percentage);
		exit(2);
	}
	if( percentage >= warncpu ) {
		printf( "WARN: CPU usage is %.2f%% of allocated maximum\n",percentage);
		exit(1);
	}
	printf( "OK: CPU usage is %.2f%% of allocated maximum\n",percentage);
	exit(0);
}
