/* check_vmem
	Nagios plugin for Linux guest running under VMWARE ESX3
	Steve Shipway, University of Auckland
 
	Incorporates some example code as shipped with SDK

	Requires VMware Tools and SDK
*/

/* Includes as given */
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <math.h>
#ifdef _WIN32
#	include <windows.h>
#else
#	include <unistd.h>
#	include <dlfcn.h>
#endif
#include "vmGuestLib.h"
#ifdef WIN32
#define SLEEP(x) Sleep(x * 1000)
#define TMPFILE "C:\\vmware.txt"
#else
#define SLEEP(x) sleep(x)
#define TMPFILE "/tmp/vmware-data"
#endif

static Bool done = FALSE;
/* Functions to dynamically load from the GuestLib library. */
char const * (*GuestLib_GetErrorText)(VMGuestLibError);
VMGuestLibError (*GuestLib_OpenHandle)(VMGuestLibHandle*);
VMGuestLibError (*GuestLib_CloseHandle)(VMGuestLibHandle);
VMGuestLibError (*GuestLib_UpdateInfo)(VMGuestLibHandle handle);
VMGuestLibError (*GuestLib_GetSessionId)(VMGuestLibHandle handle,
				                             VMSessionId *id);
VMGuestLibError (*GuestLib_GetMemReservationMB)(VMGuestLibHandle handle,
				                                    uint32 *memReservationMB);
VMGuestLibError (*GuestLib_GetMemLimitMB)(VMGuestLibHandle handle, uint32 *memLimitMB);
VMGuestLibError (*GuestLib_GetMemShares)(VMGuestLibHandle handle, uint32 *memShares);
VMGuestLibError (*GuestLib_GetMemMappedMB)(VMGuestLibHandle handle,
				                               uint32 *memMappedMB);
VMGuestLibError (*GuestLib_GetMemActiveMB)(VMGuestLibHandle handle, uint32 *memActiveMB);
VMGuestLibError (*GuestLib_GetMemOverheadMB)(VMGuestLibHandle handle,
				                                 uint32 *memOverheadMB);
VMGuestLibError (*GuestLib_GetMemBalloonedMB)(VMGuestLibHandle handle,
				                                  uint32 *memBalloonedMB);
VMGuestLibError (*GuestLib_GetMemSwappedMB)(VMGuestLibHandle handle,
				                                uint32 *memSwappedMB);
VMGuestLibError (*GuestLib_GetMemSharedMB)(VMGuestLibHandle handle,
				                               uint32 *memSharedMB);
VMGuestLibError (*GuestLib_GetMemSharedSavedMB)(VMGuestLibHandle handle,
				                                    uint32 *memSharedSavedMB);
VMGuestLibError (*GuestLib_GetMemUsedMB)(VMGuestLibHandle handle,
				                             uint32 *memUsedMB);

/*
 * Handle for use with shared library.
 */

#ifdef WIN32
HMODULE dlHandle = NULL;
#else
void *dlHandle = NULL;
#endif

/*
 * GuestLib handle.
 */
VMGuestLibHandle glHandle;


/*
 * Macro to load a single GuestLib function from the shared library.
 */

#ifdef _WIN32
#define LOAD_ONE_FUNC(funcname)				                          \
	do {			                                                     \
		(FARPROC)funcname = GetProcAddress(dlHandle, "VM" #funcname);  \
		if (funcname == NULL) {		                                  \
			error = GetLastError();	                                  \
			printf("Failed to load \'%s\': %d\n",	                    \
				    #funcname, error);                                   \
			return FALSE;	                                            \
		}		                                                        \
	} while (0)

#else

#define LOAD_ONE_FUNC(funcname)				               \
	do {			                                          \
		funcname = dlsym(dlHandle, "VM" #funcname);		   \
		if ((dlErrStr = dlerror()) != NULL) {		         \
			printf("Failed to load \'%s\': \'%s\'\n",	     \
				    #funcname, dlErrStr);                     \
			return FALSE;	                                 \
		}		                                             \
	} while (0)

#endif

/*
 *-----------------------------------------------------------------------------
 *
 * LoadFunctions --
 *
 *		Load the functions from the shared library.
 *
 * Results:
 *		TRUE on success
 *		FALSE on failure
 *
 * Side effects:
 *		None
 *
 *-----------------------------------------------------------------------------
 */

Bool LoadFunctions(void) {
	/*
	 * First, try to load the shared library.
	 */
#ifdef _WIN32
	DWORD error;

	dlHandle = LoadLibrary("vmGuestLib.dll");
	if (!dlHandle) {
		error = GetLastError();
		printf("LoadLibrary failed: %d\n", error);
		return FALSE;
	}
#else
	char const *dlErrStr;

	dlHandle = dlopen("libvmGuestLib.so", RTLD_NOW);
	if (!dlHandle) {
		dlErrStr = dlerror();
		printf("dlopen failed: \'%s\'\n", dlErrStr);
		return FALSE;
	}
#endif

	/* Load all the individual library functions. */
	LOAD_ONE_FUNC(GuestLib_GetErrorText);
	LOAD_ONE_FUNC(GuestLib_OpenHandle);
	LOAD_ONE_FUNC(GuestLib_CloseHandle);
	LOAD_ONE_FUNC(GuestLib_UpdateInfo);
	LOAD_ONE_FUNC(GuestLib_GetSessionId);
	LOAD_ONE_FUNC(GuestLib_GetMemReservationMB);
	LOAD_ONE_FUNC(GuestLib_GetMemLimitMB);
	LOAD_ONE_FUNC(GuestLib_GetMemShares);
	LOAD_ONE_FUNC(GuestLib_GetMemMappedMB);
	LOAD_ONE_FUNC(GuestLib_GetMemActiveMB);
	LOAD_ONE_FUNC(GuestLib_GetMemOverheadMB);
	LOAD_ONE_FUNC(GuestLib_GetMemBalloonedMB);
	LOAD_ONE_FUNC(GuestLib_GetMemSwappedMB);
	LOAD_ONE_FUNC(GuestLib_GetMemSharedMB);
	LOAD_ONE_FUNC(GuestLib_GetMemSharedSavedMB);
	LOAD_ONE_FUNC(GuestLib_GetMemUsedMB);

	return TRUE;
}


/* Actually obtain the VMWare ESX statistics */
Bool getdata(void) {
	VMGuestLibError glError;
	Bool success = TRUE;
	uint32 memReservationMB = 0;
	uint32 memLimitMB = 0;
	uint32 memShares = 0;
	uint32 memMappedMB = 0;
	uint32 memActiveMB = 0;
	uint32 memOverheadMB = 0;
	uint32 memBalloonedMB = 0;
	uint32 memSwappedMB = 0;
	uint32 memSharedMB = 0;
	uint32 memSharedSavedMB = 0;
	uint32 memUsedMB = 0;
	uint64 elapsedMs = 0;
	VMSessionId sessionId = 0;

	/* Try to load the library. */
	glError = GuestLib_OpenHandle(&glHandle);
	if (glError != VMGUESTLIB_ERROR_SUCCESS) {
		printf("UNKNOWN: Is vmware-tools installed correctly?  OpenHandle failed: %s\n", GuestLib_GetErrorText(glError));
	exit(3);
	}

	/* Attempt to retrieve info from the host. */
		VMSessionId tmpSession;

		glError = GuestLib_UpdateInfo(glHandle);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("UNKNOWN: Is vmware-tools working correctly?  UpdateInfo failed: %s\n", GuestLib_GetErrorText(glError));
			exit(3);
		}
		/* Retrieve and check the session ID */
		glError = GuestLib_GetSessionId(glHandle, &tmpSession);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("UNKNOWN: Failed to get session ID: %s\n", GuestLib_GetErrorText(glError));	
			exit(3);
		}

		if (tmpSession == 0) {
			printf("UNKNOWN: Got zero sessionId from GuestLib.  Do you have the latest version of vmware-tools?\n");
			exit(3);
		}
		sessionId = tmpSession;
		/* Retrieve all the stats. */
		glError = GuestLib_GetMemReservationMB(glHandle, &memReservationMB);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("Failed to get mem reservation: %s\n",
				    GuestLib_GetErrorText(glError));
			exit(3); }
		glError = GuestLib_GetMemLimitMB(glHandle, &memLimitMB);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("Failed to get mem limit: %s\n", GuestLib_GetErrorText(glError));
			exit(3); }
		glError = GuestLib_GetMemShares(glHandle, &memShares);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			if (glError == VMGUESTLIB_ERROR_NOT_AVAILABLE) {
				memShares = 0;
			} else {
				printf("Failed to get mem shares: %s\n", GuestLib_GetErrorText(glError));
				exit(3);
			}
		}
		glError = GuestLib_GetMemMappedMB(glHandle, &memMappedMB);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("Failed to get mapped mem: %s\n", GuestLib_GetErrorText(glError));
			exit(3); }
		glError = GuestLib_GetMemActiveMB(glHandle, &memActiveMB);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("Failed to get active mem: %s\n", GuestLib_GetErrorText(glError));
			exit(3); }
		glError = GuestLib_GetMemOverheadMB(glHandle, &memOverheadMB);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("Failed to get overhead mem: %s\n", GuestLib_GetErrorText(glError));
			exit(3); }
		glError = GuestLib_GetMemBalloonedMB(glHandle, &memBalloonedMB);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("Failed to get ballooned mem: %s\n", GuestLib_GetErrorText(glError));
			exit(3); }
		glError = GuestLib_GetMemSwappedMB(glHandle, &memSwappedMB);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("Failed to get swapped mem: %s\n", GuestLib_GetErrorText(glError));
			exit(3); }
		glError = GuestLib_GetMemSharedMB(glHandle, &memSharedMB);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("Failed to get shared mem: %s\n", GuestLib_GetErrorText(glError));
			exit(3); }
		glError = GuestLib_GetMemSharedSavedMB(glHandle, &memSharedSavedMB);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("Failed to get saved mem: %s\n", GuestLib_GetErrorText(glError));
			exit(3); }
		glError = GuestLib_GetMemUsedMB(glHandle, &memUsedMB);
		if (glError != VMGUESTLIB_ERROR_SUCCESS) {
			printf("Failed to get used mem: %s\n", GuestLib_GetErrorText(glError));
			exit(3); }

		/* Now we have the data, we can report on it.
			We could report on memory or on CPU.  Default to CPU. */


	glError = GuestLib_CloseHandle(glHandle);
/*
	if (glError != VMGUESTLIB_ERROR_SUCCESS) {
		printf("UNKNOWN: Failed to CloseHandle: %s\n", GuestLib_GetErrorText(glError));
		exit(3);
	}
*/
	return TRUE;
}

int main(int argc, char *argv[]) {
	/* Try to load the library. */
	if (!LoadFunctions()) {
		printf("UNKNOWN: Failed to load vmware-tools library.  Are you sure that this is host is an ESX3 Guest and vmware-tools is installed?\n");
		exit(3);
	}

	/* extract the performance metrics from vmware-tools */
	if (!getdata()) {
		printf("UNKNOWN: Retrieving data from ESX failed\n");
		exit(3);
	}

#ifdef _WIN32
	if (!FreeLibrary(dlHandle)) {
		DWORD error = GetLastError();
		printf("UNKNOWN: Failed to FreeLibrary: %d\n", error);
		exit(3);
	}
#else
	if (dlclose(dlHandle)) {
		printf("UNKNOWN: dlclose failed for some reason\n");
		exit(3);
	}
#endif

	/* now we have the data!  We can do thresholding, and all sorts of 
	   other groovy stuff */
	printf( "OK: Memory usage is: %.2f%% pvt/shr/bal/swp=[%.2f%%/%.2f%%/%.2f%%/%.2f%%], %.2f%% active\n",percentage);
	exit(0);
}
